import json
import os
from pathlib import Path
import re
from typing import Tuple

import peewee

from const.messages import CHECK_AND_CONTINUE, AFFIRMATIVE_ANSWERS, NEGATIVE_ANSWERS
from utils.style import color_yellow_bold, color_cyan, color_white_bold, color_green
from const.common import STEPS
from database.database import delete_unconnected_steps_from, delete_all_app_development_data, update_app_status
from const.ipc import MESSAGE_TYPE
from prompts.prompts import ask_user
from helpers.exceptions.TokenLimitError import TokenLimitError
from utils.questionary import styled_text
from helpers.files import get_directory_contents, get_file_contents, clear_directory, update_file
from helpers.cli import build_directory_tree
from helpers.agents.TechLead import TechLead
from helpers.agents.Developer import Developer
from helpers.agents.Architect import Architect
from helpers.agents.ProductOwner import ProductOwner
from helpers.agents.TechnicalWriter import TechnicalWriter

from database.models.development_steps import DevelopmentSteps
from database.models.file_snapshot import FileSnapshot
from database.models.files import File
from logger.logger import logger
from utils.dot_gpt_pilot import DotGptPilot
from utils.llm_connection import test_api_access
from utils.ignore import IgnoreMatcher

from utils.telemetry import telemetry
from utils.task import Task

class Project:
    def __init__(
        self,
        args,
        *,
        ipc_client_instance=None,
    ):
        """
        Initialize a project.

        Args:
            args (dict): Project arguments - app_id, (app_type, name), user_id, email, password, step
            name (str, optional): Project name. Default is None.
            description (str, optional): Project description. Default is None.
            user_stories (list, optional): List of user stories. Default is None.
            user_tasks (list, optional): List of user tasks. Default is None.
            architecture (str, optional): Project architecture. Default is None.
            development_plan (str, optional): Development plan. Default is None.
            current_step (str, optional): Current step in the project. Default is None.
        """
        self.args = args
        self.llm_req_num = 0
        self.command_runs_count = 0
        self.user_inputs_count = 0
        self.current_task = Task()
        self.checkpoints = {
            'last_user_input': None,
            'last_command_run': None,
            'last_development_step': None,
        }
        # TODO make flexible
        self.root_path = ''
        self.skip_until_dev_step = None
        self.skip_steps = None
        self.main_prompt = None
        self.files = []
        self.continuing_project = args.get('continuing_project', False)

        self.ipc_client_instance = ipc_client_instance

        # self.restore_files({dev_step_id_to_start_from})

        self.finished = False
        self.current_step = None
        self.name = None
        self.project_description = None
        self.clarifications = None
        self.user_stories = None
        self.user_tasks = None
        self.architecture = None
        self.system_dependencies = None
        self.package_dependencies = None
        self.development_plan = None
        self.dot_pilot_gpt = DotGptPilot(log_chat_completions=True)

        if os.getenv("AUTOFIX_FILE_PATHS", "").lower() in ["true", "1", "yes"]:
            File.update_paths()

    def set_root_path(self, root_path: str):
        self.root_path = root_path
        self.dot_pilot_gpt.with_root_path(root_path)

    def start(self):
        """
        Start the project.
        """

        telemetry.start()
        telemetry.set("app_id", self.args["app_id"])

        if not test_api_access(self):
            return False

        self.project_manager = ProductOwner(self)
        self.project_manager.get_project_description()

        self.project_manager.get_user_stories()
        # self.user_tasks = self.project_manager.get_user_tasks()

        self.architect = Architect(self)
        self.architect.get_architecture()

        self.developer = Developer(self)
        self.developer.set_up_environment()
        self.technical_writer = TechnicalWriter(self)

        self.tech_lead = TechLead(self)
        self.tech_lead.create_development_plan()

        telemetry.set("architecture", {
            "description": self.architecture,
            "system_dependencies": self.system_dependencies,
            "package_dependencies": self.package_dependencies,
        })
        # TODO move to constructor eventually
        if self.args['step'] is not None and STEPS.index(self.args['step']) < STEPS.index('coding'):
            clear_directory(self.root_path)
            delete_all_app_development_data(self.args['app_id'])
            self.finish_loading()

        if 'skip_until_dev_step' in self.args:
            self.skip_until_dev_step = self.args['skip_until_dev_step']
            if self.args['skip_until_dev_step'] == '0':
                clear_directory(self.root_path)
                delete_all_app_development_data(self.args['app_id'])
                self.finish_loading()
            elif self.skip_until_dev_step is not None:
                should_overwrite_files = None
                while should_overwrite_files is None or should_overwrite_files.lower() not in AFFIRMATIVE_ANSWERS + NEGATIVE_ANSWERS:
                    print('yes/no', type='buttons-only')
                    should_overwrite_files = styled_text(
                        self,
                        "Can I overwrite any changes that you might have made to the project since last running GPT Pilot (y/n)?",
                        ignore_user_input_count=True
                    )

                    logger.info('should_overwrite_files: %s', should_overwrite_files)
                    if should_overwrite_files in NEGATIVE_ANSWERS:
                        break
                    elif should_overwrite_files in AFFIRMATIVE_ANSWERS:
                        FileSnapshot.delete().where(
                            FileSnapshot.app == self.app and FileSnapshot.development_step == self.skip_until_dev_step).execute()
                        self.save_files_snapshot(self.skip_until_dev_step)
                        break
        # TODO END

        self.dot_pilot_gpt.write_project(self)
        print(json.dumps({
            "project_stage": "coding"
        }), type='info')
        self.developer.start_coding()
        return True

    def finish(self):
        """
        Finish the project.
        """
        while True:
            feature_description = ask_user(self, "Project is finished! Do you want to add any features or changes? "
                                                 "If yes, describe it here and if no, just press ENTER",
                                           require_some_input=False)

            if feature_description == '':
                return

            self.tech_lead.create_feature_plan(feature_description)
            self.developer.start_coding()
            self.tech_lead.create_feature_summary(feature_description)

    def get_directory_tree(self, with_descriptions=False):
        """
        Get the directory tree of the project.

        Args:
            with_descriptions (bool, optional): Whether to include descriptions. Default is False.

        Returns:
            dict: The directory tree.
        """
        return build_directory_tree(self.root_path)

    def get_test_directory_tree(self):
        """
        Get the directory tree of the tests.

        Returns:
            dict: The directory tree of tests.
        """
        # TODO remove hardcoded path
        return build_directory_tree(self.root_path + '/tests')

    def get_all_coded_files(self):
        """
        Get all coded files in the project.

        Returns:
            list: A list of coded files.
        """
        files = (
            File
                .select()
                .where(
                    (File.app_id == self.args['app_id']) &
                    peewee.fn.EXISTS(FileSnapshot.select().where(FileSnapshot.file_id == File.id))
                )
            )

        return self.get_files([file.path + '/' + file.name for file in files])

    def get_files(self, files):
        """
        Get file contents.

        Args:
            files (list): List of file paths.

        Returns:
            list: A list of files with content.
        """
        matcher = IgnoreMatcher(root_path=self.root_path)
        files_with_content = []
        for file_path in files:
            try:
                # TODO path is sometimes relative and sometimes absolute - fix at one point
                _, full_path = self.get_full_file_path(file_path, file_path)
                file_data = get_file_contents(full_path, self.root_path)
            except ValueError:
                full_path = None
                file_data = {"path": file_path, "name": os.path.basename(file_path), "content": ''}

            if full_path and file_data["content"] != "" and not matcher.ignore(full_path):
                files_with_content.append(file_data)

        return files_with_content

    def find_input_required_lines(self, file_content):
        """
        Parses the provided string (representing file content) and returns a list of tuples containing
        the line number and line content for lines that contain the text 'INPUT_REQUIRED'.

        :param file_content: The string content of the file.
        :return: A list of tuples (line number, line content).
        """
        lines_with_input_required = []
        lines = file_content.split('\n')

        for line_number, line in enumerate(lines, start=1):
            if 'INPUT_REQUIRED' in line:
                lines_with_input_required.append((line_number, line.strip()))

        return lines_with_input_required

    def save_file(self, data):
        """
        Save a file.

        Args:
            data: { name: 'hello.py', path: 'path/to/hello.py', content: 'print("Hello!")' }
        """

        name = data['name'] if 'name' in data and data['name'] != '' else os.path.basename(data['path'])
        path = data['path'] if 'path' in data else name

        path, full_path = self.get_full_file_path(path, name)
        update_file(full_path, data['content'], project=self)
        if full_path not in self.files:
            self.files.append(full_path)

        (File.insert(app=self.app, path=path, name=name, full_path=full_path)
         .on_conflict(
            conflict_target=[File.app, File.name, File.path],
            preserve=[],
            update={'name': name, 'path': path, 'full_path': full_path})
         .execute())

        if not self.skip_steps:
            inputs_required = self.find_input_required_lines(data['content'])
            for line_number, line_content in inputs_required:
                user_input = None
                print(color_yellow_bold(f'Input required on line {line_number}:\n{line_content}') + '\n')
                while user_input is None or user_input.lower() not in AFFIRMATIVE_ANSWERS + ['continue']:
                    print({'path': full_path, 'line': line_number}, type='openFile')
                    print('continue', type='buttons-only')
                    user_input = styled_text(
                        self,
                        f'Please open the file {data["path"]} on the line {line_number} and add the required input. Once you\'re done, type "y" to continue.',
                        ignore_user_input_count=True
                    )

    def get_full_file_path(self, file_path: str, file_name: str) -> Tuple[str, str]:
        """
        Combine file path and name into a full file path.

        :param file_path: File path.
        :param file_name: File name.
        :return: (file_path, absolute_path) pair.

        Tries to combine the two in a way that makes most sense, even if the given path
        have some shared components.
        """
        def normalize_path(path: str) -> Tuple[str, str]:
            """
            Normalizes a path (see rules in comments) and returns (directory, basename) pair.

            :param path: Path to normalize.
            :return: (directory, basename) pair.

            Directory component may be empty if the path is considered to be a
            file name. Basename component may be empty if the path is considered
            to be a directory name.
            """

            # Normalize path to use os-specific separator (as GPT may output paths
            # with / even if we're on Windows)
            path = str(Path(path))

            # If a path references user's home directory (~), we only care about
            # the relative part within it (assume ~ is meant to be the project path).
            # Examples:
            # - /Users/zvonimirsabljic/Development/~/pilot/server.js -> /pilot/server.js
            # - ~/pilot/server.js -> /pilot/server.js
            if "~" in path:
                path = path.split("~")[-1]

            # If the path explicitly references the current directory, remove it so we
            # can nicely use it for joins later.
            if path == "." or path.startswith(f".{os.path.sep}"):
                path = path[1:]

            # If the path is absolute, we only care about the relative part within
            # the project directory (assume the project directory is the root).
            # Examples:
            # - /Users/zvonimirsabljic/Development/copilot/pilot/server.js -> /pilot/server.js
            # - /pilot/server.js -> /pilot/server.js
            # - C:\Users\zvonimirsabljic\Development\copilot\pilot\server.js -> \pilot\server.js
            path = path.replace(self.root_path, '')

            # If the final component of the path doesn't have a file extension,
            # assume it's a directory and add a final (back)slash.
            # Examples:
            # - /pilot/server.js -> /pilot/server.js
            # - /pilot -> /pilot/
            # - \pilot\server.js -> \pilot\server.js
            # - \pilot -> \pilot\
            base = os.path.basename(path)
            if base and "." not in base:
                path += os.path.sep

            # In case we're in Windows and dealing with full paths, remove the drive letter.
            _, path = os.path.splitdrive(path)

            # We want all paths to start with / (or \\ in Windows)
            if not path.startswith(os.path.sep):
                path = os.path.sep + path

            return os.path.split(path)

        head_path, tail_path = normalize_path(file_path)
        head_name, tail_name = normalize_path(file_name)

        # Prefer directory path from the first argument (file_path), and
        # prefer the file name from the second argument (file_name).
        final_file_path = head_path if head_path != '' else head_name
        final_file_name = tail_name if tail_name != '' else tail_path

        # If the directory is contained in the second argument (file_name),
        # use that (as it might include additional subdirectories).
        if head_path in head_name:
            final_file_path = head_name

        # Try to combine the directory and file name from the two arguments
        # in the way that makes the most sensible output.
        if final_file_path != head_name and head_name not in head_path:
            if '.' in tail_path:
                final_file_path = head_name + head_path
            else:
                final_file_path = head_path + head_name

        if final_file_path == '':
            final_file_path = os.path.sep

        final_absolute_path = os.path.join(self.root_path, final_file_path[1:], final_file_name)
        return final_file_path, final_absolute_path


    def save_files_snapshot(self, development_step_id):
        files = get_directory_contents(self.root_path)
        development_step, created = DevelopmentSteps.get_or_create(id=development_step_id)

        total_files = 0
        total_lines = 0

        for file in files:
            if not self.check_ipc():
                print(color_cyan(f'Saving file {file["full_path"]}'))
            # TODO this can be optimized so we don't go to the db each time
            file_in_db, created = File.get_or_create(
                app=self.app,
                name=file['name'],
                path=file['path'],
                full_path=file['full_path'],
            )

            file_snapshot, created = FileSnapshot.get_or_create(
                app=self.app,
                development_step=development_step,
                file=file_in_db,
                defaults={'content': file.get('content', '')}
            )
            file_snapshot.content = file['content']
            file_snapshot.save()
            total_files += 1
            if isinstance(file['content'], str):
                total_lines += file['content'].count('\n') + 1

        telemetry.set("num_files", total_files)
        telemetry.set("num_lines", total_lines)

    def restore_files(self, development_step_id):
        development_step = DevelopmentSteps.get(DevelopmentSteps.id == development_step_id)
        file_snapshots = FileSnapshot.select().where(FileSnapshot.development_step == development_step)

        clear_directory(self.root_path, ignore=self.files)
        for file_snapshot in file_snapshots:
            update_file(file_snapshot.file.full_path, file_snapshot.content, project=self)
            if file_snapshot.file.full_path not in self.files:
                self.files.append(file_snapshot.file.full_path)

    def delete_all_steps_except_current_branch(self):
        delete_unconnected_steps_from(self.checkpoints['last_development_step'], 'previous_step')
        delete_unconnected_steps_from(self.checkpoints['last_command_run'], 'previous_step')
        delete_unconnected_steps_from(self.checkpoints['last_user_input'], 'previous_step')

    def ask_for_human_intervention(self, message, description=None, cbs={}, convo=None, is_root_task=False):
        answer = ''
        question = color_yellow_bold(message)

        if description is not None:
            question += '\n' + '-' * 100 + '\n' + color_white_bold(description) + '\n' + '-' * 100 + '\n'

        reset_branch_id = None if convo is None else convo.save_branch()

        while answer.lower() != 'continue':
            print('continue', type='button')
            answer = ask_user(self, CHECK_AND_CONTINUE,
                              require_some_input=False,
                              hint=question)

            try:
                if answer.lower() in cbs:
                    return cbs[answer.lower()](convo)
                elif answer != '':
                    return {'user_input': answer}
            except TokenLimitError as e:
                if is_root_task and answer.lower() not in cbs and answer != '':
                    convo.load_branch(reset_branch_id)
                    return {'user_input': answer}
                else:
                    raise e

    def log(self, text, message_type):
        if self.check_ipc():
            self.ipc_client_instance.send({
                'type': MESSAGE_TYPE[message_type],
                'content': str(text),
            })
            if message_type == MESSAGE_TYPE['user_input_request']:
                return self.ipc_client_instance.listen()
        else:
            print(text)

    def check_ipc(self):
        """
        Checks if there is an open Inter-Process Communication (IPC) connection.

        Returns:
            bool: True if there is an open IPC connection, False otherwise.
        """
        return self.ipc_client_instance is not None and self.ipc_client_instance.client is not None

    def finish_loading(self):
        print('', type='loadingFinished')
        self.skip_steps = False



import os
from unittest.mock import patch
from utils.files import setup_workspace


def mocked_create_directory(path, exist_ok=True):
    return


def mocked_abspath(file):
    return "/root_path/pilot/helpers"


@patch('utils.files.os.makedirs', side_effect=mocked_create_directory)
def test_setup_workspace_with_existing_workspace(mock_makedirs):
    args = {'workspace': '/some/directory', 'name': 'sample'}
    result = setup_workspace(args)
    assert result == '/some/directory'


def test_setup_workspace_with_root_arg(monkeypatch):
    args = {'root': '/my/root', 'name': 'project_name'}

    monkeypatch.setattr('os.path.abspath', mocked_abspath)
    monkeypatch.setattr('os.makedirs', mocked_create_directory)

    result = setup_workspace(args)
    assert result.replace('\\', '/') == "/my/root/workspace/project_name"


@patch('utils.files.os.path.abspath', return_value='/root_path/pilot/helpers')
@patch('utils.files.os.makedirs', side_effect=mocked_create_directory)
def test_setup_workspace_without_existing_workspace(mock_makedirs, mock_abs_path):
    args = {'workspace': None, 'name': 'project_name'}

    result = setup_workspace(args)
    assert result.replace('\\', '/') == "/root_path/workspace/project_name"



import platform
import uuid
import re

from const.code_execution import MAX_COMMAND_DEBUG_TRIES, MAX_RECUSION_LAYER
from const.function_calls import DEBUG_STEPS_BREAKDOWN
from const.messages import AFFIRMATIVE_ANSWERS, NEGATIVE_ANSWERS
from helpers.AgentConvo import AgentConvo
from helpers.exceptions.TokenLimitError import TokenLimitError
from helpers.exceptions.TooDeepRecursionError import TooDeepRecursionError
from logger.logger import logger
from prompts.prompts import ask_user


class Debugger:
    def __init__(self, agent):
        self.agent = agent
        self.recursion_layer = 0

    def debug(self, convo, command=None, user_input=None, issue_description=None, is_root_task=False, ask_before_debug=False):
        """
        Debug a conversation.

        Args:
            convo (AgentConvo): The conversation object.
            command (dict, optional): The command to debug. Default is None.
            user_input (str, optional): User input for debugging. Default is None.
                Should provide `command` or `user_input`.
            issue_description (str, optional): Description of the issue to debug. Default is None.
            ask_before_debug (bool, optional): True if we have to ask user for permission to start debugging.

        Returns:
            bool: True if debugging was successful, False otherwise.
        """
        logger.info('Debugging %s', command)
        self.recursion_layer += 1
        self.agent.project.current_task.add_debugging_task(self.recursion_layer, command, user_input, issue_description)
        if self.recursion_layer > MAX_RECUSION_LAYER:
            self.recursion_layer = 0
            raise TooDeepRecursionError()

        function_uuid = str(uuid.uuid4())
        convo.save_branch(function_uuid)
        success = False

        for i in range(MAX_COMMAND_DEBUG_TRIES):
            if success:
                break

            if ask_before_debug or i > 0:
                print('yes/no', type='button')
                answer = ask_user(self.agent.project, 'Can I start debugging this issue [Y/n/error details]?', require_some_input=False)
                if answer.lower() in NEGATIVE_ANSWERS:
                    return True
                if answer and answer.lower() not in AFFIRMATIVE_ANSWERS:
                    user_input = answer
                    self.agent.project.current_task.add_user_input_to_debugging_task(user_input)

            llm_response = convo.send_message('dev_ops/debug.prompt',
                {
                    'command': command['command'] if command is not None else None,
                    'user_input': user_input,
                    'issue_description': issue_description,
                    'os': platform.system(),
                    'context': convo.to_context_prompt()
                },
                DEBUG_STEPS_BREAKDOWN)

            completed_steps = []

            try:
                while True:
                    logger.info('Thoughts: ' + llm_response['thoughts'])
                    logger.info('Reasoning: ' + llm_response['reasoning'])
                    steps = completed_steps + llm_response['steps']

                    # TODO refactor to nicely get the developer agent
                    result = self.agent.project.developer.execute_task(
                        convo,
                        f"Thoughts: {llm_response['thoughts']}\n\nReasoning: {llm_response['reasoning']}",
                        steps,
                        test_command=command,
                        test_after_code_changes=True,
                        continue_development=False,
                        is_root_task=is_root_task,
                        continue_from_step=len(completed_steps)
                    )

                    # in case one step failed or llm wants to see the output to determine the next steps
                    if 'step_index' in result:
                        result['os'] = platform.system()
                        step_index = result['step_index']
                        completed_steps = steps[:step_index+1]
                        result['completed_steps'] = completed_steps
                        result['current_step'] = steps[step_index]
                        result['next_steps'] = steps[step_index + 1:]
                        result['current_step_index'] = step_index

                        # Remove the previous debug plan and build a new one
                        convo.remove_last_x_messages(2)
                        # todo before updating task first check if update is needed
                        llm_response = convo.send_message('development/task/update_task.prompt', result,
                            DEBUG_STEPS_BREAKDOWN)
                    else:
                        success = result['success']
                        if not success:
                            convo.load_branch(function_uuid)
                            if 'cli_response' in result:
                                user_input = result['cli_response']
                                convo.messages[-2]['content'] = re.sub(
                                    r'(?<=The output was:\n\n).*?(?=\n\nThink about this output)',
                                    AgentConvo.escape_specials(result['cli_response']),
                                    convo.messages[-2]['content'],
                                    flags=re.DOTALL
                                )
                        break

            except TokenLimitError as e:
                if self.recursion_layer > 0:
                    self.recursion_layer -= 1
                    raise e
                else:
                    if not success:
                        convo.load_branch(function_uuid)
                    continue

            # if not success:
            #     # TODO explain better how should the user approach debugging
            #     # we can copy the entire convo to clipboard so they can paste it in the playground
            #     user_input = convo.agent.project.ask_for_human_intervention(
            #         'It seems like I cannot debug this problem by myself. Can you please help me and try debugging it yourself?' if user_input is None else f'Can you check this again:\n{issue_description}?',
            #         response['data']
            #     )

            #     if user_input == 'continue':
            #         success = True

        self.recursion_layer -= 1
        return success



from const.function_calls import ARCHITECTURE, USER_TASKS
from utils.llm_connection import clean_json_response
from .function_calling import parse_agent_response, JsonPrompter


class TestFunctionCalling:
    def test_parse_agent_response_text(self):
        # Given
        response = {'text': 'Hello world!'}

        # When
        response = parse_agent_response(response, None)

        # Then
        assert response == 'Hello world!'

    def test_parse_agent_response_json(self):
        # Given
        response = {'text': '{"greeting": "Hello world!"}'}
        function_calls = {'definitions': [], 'functions': {}}

        # When
        response = parse_agent_response(response, function_calls)

        # Then
        assert response == {'greeting': 'Hello world!'}

    def test_parse_agent_response_json_markdown(self):
        # Given
        response = {'text': '```json\n{"greeting": "Hello world!"}\n```'}
        function_calls = {'definitions': [], 'functions': {}}

        # When
        response['text'] = clean_json_response(response['text'])
        response = parse_agent_response(response, function_calls)

        # Then
        assert response == {'greeting': 'Hello world!'}

    def test_parse_agent_response_markdown(self):
        # Given
        response = {'text': '```\n{"greeting": "Hello world!"}\n```'}
        function_calls = {'definitions': [], 'functions': {}}

        # When
        response['text'] = clean_json_response(response['text'])
        response = parse_agent_response(response, function_calls)

        # Then
        assert response == {'greeting': 'Hello world!'}

    def test_parse_agent_response_multiple_args(self):
        # Given
        response = {'text': '{"greeting": "Hello", "name": "John"}'}
        function_calls = {'definitions': [], 'functions': {}}

        # When
        response = parse_agent_response(response, function_calls)

        # Then
        assert response['greeting'] == 'Hello'
        assert response['name'] == 'John'


def test_json_prompter():
    # Given
    prompter = JsonPrompter()

    # When
    prompt = prompter.prompt('Create a web-based chat app', ARCHITECTURE['definitions'])  # , 'process_technologies')

    # Then
    assert prompt == '''Help choose the appropriate function to call to answer the user's question.
You must respond with ONLY the JSON object, with NO additional text or explanation.

Available functions:
- process_architecture - Get architecture and the list of system dependencies required for the project.

Create a web-based chat app'''


def test_llama_json_prompter():
    # Given
    prompter = JsonPrompter(is_instruct=True)

    # When
    prompt = prompter.prompt('Create a web-based chat app', ARCHITECTURE['definitions'])  # , 'process_technologies')

    # Then
    assert prompt == '''[INST] <<SYS>>
Help choose the appropriate function to call to answer the user's question.
You must respond with ONLY the JSON object, with NO additional text or explanation.

Available functions:
- process_architecture - Get architecture and the list of system dependencies required for the project.
<</SYS>>

Create a web-based chat app [/INST]'''


def test_json_prompter_named():
    # Given
    prompter = JsonPrompter()

    # When
    prompt = prompter.prompt('Create a web-based chat app', USER_TASKS['definitions'], 'process_user_tasks')

    # Then
    assert prompt == '''**IMPORTANT**
You must respond with ONLY the JSON object, with NO additional text or explanation.

Here is the schema for the expected JSON object:
```json
{
    "tasks": {
        "type": "array",
        "description": "List of user tasks.",
        "items": {
            "type": "string",
            "description": "user task"
        }
    }
}
```

Create a web-based chat app'''


def test_llama_json_prompter_named():
    # Given
    prompter = JsonPrompter(is_instruct=True)

    # When
    prompt = prompter.prompt('Create a web-based chat app', USER_TASKS['definitions'], 'process_user_tasks')

    # Then
    assert prompt == '''[INST] <<SYS>>
**IMPORTANT**
You must respond with ONLY the JSON object, with NO additional text or explanation.

Here is the schema for the expected JSON object:
```json
{
    "tasks": {
        "type": "array",
        "description": "List of user tasks.",
        "items": {
            "type": "string",
            "description": "user task"
        }
    }
}
```
<</SYS>>

Create a web-based chat app [/INST]'''



# from .run_command_until_success import run_command_until_success
from .cli_execute_command import cli_execute_command
from .Dev_continue_development import test_continue_development
from .utils import use_args


def run_test(test_name: str, args):
    print(f'Running UX test "{test_name}"...')

    tests = {
        # 'run_command_until_success': run_command_until_success,
        'cli_execute_command': cli_execute_command,
        'continue_development': test_continue_development,
    }

    if test_name in tests:
        use_args(args)
        return tests[test_name]()

    print(f'UX test "{test_name}" not found')



from playhouse.shortcuts import model_to_dict
from utils.style import color_yellow, color_red
from peewee import DoesNotExist, IntegrityError
from functools import reduce
import operator
from database.config import DB_NAME, DB_HOST, DB_PORT, DB_USER, DB_PASSWORD, DATABASE_TYPE
if DATABASE_TYPE == "postgres":
    import psycopg2
    from psycopg2.extensions import quote_ident

import os
from const.common import PROMPT_DATA_TO_IGNORE, STEPS
from logger.logger import logger
from database.models.components.base_models import database
from database.models.user import User
from database.models.app import App
from database.models.project_description import ProjectDescription
from database.models.user_stories import UserStories
from database.models.user_tasks import UserTasks
from database.models.architecture import Architecture
from database.models.development_planning import DevelopmentPlanning
from database.models.development_steps import DevelopmentSteps
from database.models.environment_setup import EnvironmentSetup
from database.models.development import Development
from database.models.file_snapshot import FileSnapshot
from database.models.command_runs import CommandRuns
from database.models.user_apps import UserApps
from database.models.user_inputs import UserInputs
from database.models.files import File
from database.models.feature import Feature

TABLES = [
            User,
            App,
            ProjectDescription,
            UserStories,
            UserTasks,
            Architecture,
            DevelopmentPlanning,
            DevelopmentSteps,
            EnvironmentSetup,
            Development,
            FileSnapshot,
            CommandRuns,
            UserApps,
            UserInputs,
            File,
            Feature,
        ]


def get_created_apps():
    return [model_to_dict(app) for app in App.select().where((App.name.is_null(False)) & (App.status.is_null(False)))]


def get_created_apps_with_steps():
    apps = get_created_apps()
    for app in apps:
        app['id'] = str(app['id'])
        app['steps'] = [step for step in STEPS[:STEPS.index(app['status']) + 1]] if app['status'] is not None else []
        app['development_steps'] = get_all_app_development_steps(app['id'])
        # TODO this is a quick way to remove the unnecessary fields from the response
        app['development_steps'] = [{k: v for k, v in dev_step.items() if k in {'id', 'created_at'}} for dev_step in
                                    app['development_steps']]
    return apps


def get_all_app_development_steps(app_id):
    return [model_to_dict(dev_step) for dev_step in DevelopmentSteps.select().where(DevelopmentSteps.app == app_id)]


def save_user(user_id, email, password):
    try:
        user = User.get(User.id == user_id)
        return user
    except DoesNotExist:
        try:
            existing_user = User.get(User.email == email)
            return existing_user
        except DoesNotExist:
            return User.create(id=user_id, email=email, password=password)


def update_app_status(app_id, new_status):
    try:
        app = App.get(App.id == app_id)
        app.status = new_status
        app.save()
        return True
    except DoesNotExist:
        return False


def get_user(user_id=None, email=None):
    if not user_id and not email:
        raise ValueError("Either user_id or email must be provided")

    query = []
    if user_id:
        query.append(User.id == user_id)
    if email:
        query.append(User.email == email)

    try:
        user = User.get(reduce(operator.or_, query))
        return user
    except DoesNotExist:
        raise ValueError("No user found with provided id or email")


def save_app(project):
    args = project.args
    app_status = getattr(project, "current_step", None)

    try:
        app = project.app
        if app is None:
            app = App.get(App.id == args['app_id'])
        for key, value in args.items():
            if key != 'app_id' and value is not None:
                setattr(app, key, value)

        app.status = app_status
        app.save()
    except DoesNotExist:
        if args.get('user_id') is not None:
            try:
                user = get_user(user_id=args['user_id'])
            except ValueError:
                user = save_user(args['user_id'], args['email'], args['password'])
                args['user_id'] = user.id
                args['email'] = user.email
        else:
            user = None

        app = App.create(
            id=args['app_id'],
            user=user,
            app_type=args.get('app_type'),
            name=args.get('name'),
            status=app_status
        )

    return app


def save_user_app(user_id, app_id, workspace):
    try:
        user_app = UserApps.get((UserApps.user == user_id) & (UserApps.app == app_id))
        user_app.workspace = workspace
        user_app.save()
    except DoesNotExist:
        user_app = UserApps.create(user=user_id, app=app_id, workspace=workspace)

    return user_app


def save_progress(app_id, step, data):
    progress_table_map = {
        'project_description': ProjectDescription,
        'user_stories': UserStories,
        'user_tasks': UserTasks,
        'architecture': Architecture,
        'development_planning': DevelopmentPlanning,
        'environment_setup': EnvironmentSetup,
        'development': Development,
    }

    data['step'] = step

    ProgressTable = progress_table_map.get(step)
    if not ProgressTable:
        raise ValueError(f"Invalid step: {step}")

    app = get_app(app_id)

    # Use the get_or_create method, which attempts to retrieve a record
    # or creates a new one if it does not exist.
    progress, created = ProgressTable.get_or_create(app=app, defaults=data)

    # If the record was not created, it already existed and we should update it
    if not created:
        for key, value in data.items():
            setattr(progress, key, value)
        progress.save()

    update_app_status(app_id, step)
    return progress


def get_app(app_id, error_if_not_found=True):
    try:
        app = App.get(App.id == app_id)
        return app
    except DoesNotExist:
        if error_if_not_found:
            raise ValueError(f"No app with id: {app_id}; use python main.py --get-created-apps-with-steps to see created apps")
        return None


def get_app_by_user_workspace(user_id, workspace):
    try:
        user_app = UserApps.get((UserApps.user == user_id) & (UserApps.workspace == workspace))
        return user_app.app
    except DoesNotExist:
        return None


def get_progress_steps(app_id, step=None):
    progress_table_map = {
        'project_description': ProjectDescription,
        'user_stories': UserStories,
        'user_tasks': UserTasks,
        'architecture': Architecture,
        'development_planning': DevelopmentPlanning,
        'environment_setup': EnvironmentSetup,
        'development': Development,
    }

    if step:
        ProgressTable = progress_table_map.get(step)
        if not ProgressTable:
            raise ValueError(f"Invalid step: {step}")

        try:
            progress = ProgressTable.get(ProgressTable.app_id == app_id)
            return model_to_dict(progress)
        except DoesNotExist:
            return None
    else:
        steps = {}
        for step, ProgressTable in progress_table_map.items():
            try:
                progress = ProgressTable.get(ProgressTable.app_id == app_id)
                steps[step] = model_to_dict(progress)
            except DoesNotExist:
                steps[step] = None

        return steps


def get_db_model_from_hash_id(model, app_id, previous_step, high_level_step):
    try:
        db_row = model.get(
            (model.app == app_id) & (model.previous_step == previous_step) & (model.high_level_step == high_level_step))
    except DoesNotExist:
        return None
    return db_row


def hash_and_save_step(Model, app_id, unique_data_fields, data_fields, message):
    # app = get_app(app_id)

    # fields_to_preserve = [getattr(Model, field) for field in list(unique_data_fields.keys())]

    for field, value in data_fields.items():
        unique_data_fields[field] = value

    try:
        # existing_record = Model.get_or_none(
        #     (Model.app == app) & (Model.previous_step == unique_data_fields['previous_step']) & (
        #                 Model.high_level_step == unique_data_fields['high_level_step']))
        inserted_id = (Model
                       .insert(**unique_data_fields)
                       .execute())

        record = Model.get_by_id(inserted_id)
        logger.debug(color_yellow(f"{message} with id {record.id}"))
    except IntegrityError:
        logger.warn(f"A record with data {unique_data_fields} already exists for {Model.__name__}.")
        return None
    return record


def save_development_step(project, prompt_path, prompt_data, messages, llm_response, exception=None):
    data_fields = {
        'messages': messages,
        'llm_response': llm_response,
        'prompt_path': prompt_path,
        'prompt_data': {} if prompt_data is None else {k: v for k, v in prompt_data.items() if
                                                       k not in PROMPT_DATA_TO_IGNORE and not callable(v)},
        'llm_req_num': project.llm_req_num,
        'token_limit_exception_raised': exception
    }

    unique_data = {
        'app': project.args['app_id'],
        'previous_step': project.checkpoints['last_development_step'],
        'high_level_step': project.current_step,
    }

    development_step = hash_and_save_step(DevelopmentSteps, project.args['app_id'], unique_data, data_fields,
                                          "Saved Development Step")
    project.checkpoints['last_development_step'] = development_step

    project.save_files_snapshot(development_step.id)

    return development_step


def get_saved_development_step(project):
    development_step = get_db_model_from_hash_id(DevelopmentSteps, project.args['app_id'],
                                                 project.checkpoints['last_development_step'], project.current_step)

    if development_step is None and project.skip_steps:
        project.finish_loading()
    return development_step


def save_command_run(project, command, cli_response, done_or_error_response, exit_code):
    if project.current_step != 'coding':
        return

    unique_data = {
        'app': project.args['app_id'],
        'previous_step': project.checkpoints['last_command_run'],
        'high_level_step': project.current_step,
    }

    data_fields = {
        'command': command,
        'cli_response': cli_response,
        'done_or_error_response': done_or_error_response,
        'exit_code': exit_code,
    }

    command_run = hash_and_save_step(CommandRuns, project.args['app_id'], unique_data, data_fields, "Saved Command Run")
    project.checkpoints['last_command_run'] = command_run
    return command_run


def get_saved_command_run(project, command):
    # data_to_hash = {
    #     'command': command,
    #     'command_runs_count': project.command_runs_count
    # }
    command_run = get_db_model_from_hash_id(CommandRuns, project.args['app_id'],
                                            project.checkpoints['last_command_run'], project.current_step)

    if command_run is None and project.skip_steps:
        project.finish_loading()
    return command_run


def save_user_input(project, query, user_input, hint):
    if project.current_step != 'coding':
        return

    unique_data = {
        'app': project.args['app_id'],
        'previous_step': project.checkpoints['last_user_input'],
        'high_level_step': project.current_step,
    }
    data_fields = {
        'query': query,
        'user_input': user_input,
        'hint': hint,
    }
    user_input = hash_and_save_step(UserInputs, project.args['app_id'], unique_data, data_fields, "Saved User Input")
    project.checkpoints['last_user_input'] = user_input
    return user_input


def get_saved_user_input(project, query):
    # data_to_hash = {
    #     'query': query,
    #     'user_inputs_count': project.user_inputs_count
    # }
    user_input = get_db_model_from_hash_id(UserInputs, project.args['app_id'], project.checkpoints['last_user_input'],
                                           project.current_step)

    if user_input is None and project.skip_steps:
        project.finish_loading()
    return user_input


def delete_all_subsequent_steps(project):
    app = get_app(project.args['app_id'])
    delete_subsequent_steps(DevelopmentSteps, app, project.checkpoints['last_development_step'])
    delete_subsequent_steps(CommandRuns, app, project.checkpoints['last_command_run'])
    delete_subsequent_steps(UserInputs, app, project.checkpoints['last_user_input'])


def delete_subsequent_steps(Model, app, step):
    logger.info(color_red(f"Deleting subsequent {Model.__name__} steps after {step.id if step is not None else None}"))
    subsequent_steps = Model.select().where(
        (Model.app == app) & (Model.previous_step == (step.id if step is not None else None)))
    for subsequent_step in subsequent_steps:
        if subsequent_step:
            delete_subsequent_steps(Model, app, subsequent_step)
            subsequent_step.delete_instance()
            if Model == DevelopmentSteps:
                FileSnapshot.delete().where(FileSnapshot.development_step == subsequent_step).execute()
                Feature.delete().where(Feature.previous_step == subsequent_step).execute()


def get_all_connected_steps(step, previous_step_field_name):
    """Recursively get all steps connected to the given step."""
    connected_steps = [step]
    prev_step = getattr(step, previous_step_field_name)
    while prev_step is not None:
        connected_steps.append(prev_step)
        prev_step = getattr(prev_step, previous_step_field_name)
    return connected_steps


def delete_all_app_development_data(app):
    models = [DevelopmentSteps, CommandRuns, UserInputs, UserApps, File, FileSnapshot]
    for model in models:
        model.delete().where(model.app == app).execute()


def delete_unconnected_steps_from(step, previous_step_field_name):
    if step is None:
        return
    connected_steps = get_all_connected_steps(step, previous_step_field_name)
    connected_step_ids = [s.id for s in connected_steps]

    unconnected_steps = DevelopmentSteps.select().where(
        (DevelopmentSteps.app == step.app) &
        (DevelopmentSteps.id.not_in(connected_step_ids))
    ).order_by(DevelopmentSteps.id.desc())

    for unconnected_step in unconnected_steps:
        print(color_red(f"Deleting unconnected {step.__class__.__name__} step {unconnected_step.id}"))
        unconnected_step.delete_instance()


def save_file_description(project, path, name, description):
    (File.insert(app=project.app, path=path, name=name, description=description)
     .on_conflict(
        conflict_target=[File.app, File.name, File.path],
        preserve=[],
        update={'description': description})
     .execute())


def save_feature(app_id, summary, messages, previous_step):
    try:
        app = get_app(app_id)
        feature = Feature.create(app=app, summary=summary, messages=messages, previous_step=previous_step)
        return feature
    except DoesNotExist:
        raise ValueError(f"No app with id: {app_id}")


def get_features_by_app_id(app_id):
    try:
        app = get_app(app_id)
        features = Feature.select().where(Feature.app == app).order_by(Feature.created_at)
        features_dict = [model_to_dict(feature) for feature in features]

        # return only 'summary' because we store all prompt_data to DB
        return [{'summary': feature['summary']} for feature in features_dict]
    except DoesNotExist:
        raise ValueError(f"No app with id: {app_id}")


def create_tables():
    with database:
        database.create_tables(TABLES)


def drop_tables():
    with database.atomic():
        for table in TABLES:
            if DATABASE_TYPE == "postgres":
                sql = f'DROP TABLE IF EXISTS "{table._meta.table_name}" CASCADE'
            elif DATABASE_TYPE == "sqlite":
                sql = f'DROP TABLE IF EXISTS "{table._meta.table_name}"'
            else:
                raise ValueError(f"Unsupported DATABASE_TYPE: {DATABASE_TYPE}")

            database.execute_sql(sql)


def database_exists():
    try:
        database.connect()
        database.close()
        return True
    except Exception:
        return False


def create_database():
    if DATABASE_TYPE == "postgres":
        # Connect to the default 'postgres' database to create a new database
        conn = psycopg2.connect(
            dbname='postgres',
            user=DB_USER,
            password=DB_PASSWORD,
            host=DB_HOST,
            port=DB_PORT
        )
        conn.autocommit = True
        cursor = conn.cursor()

        # Safely quote the database name
        safe_db_name = quote_ident(DB_NAME, conn)

        # Use the safely quoted database name in the SQL query
        cursor.execute(f"CREATE DATABASE {safe_db_name}")

        cursor.close()
        conn.close()
    else:
        pass


def tables_exist():
    if DATABASE_TYPE == "postgres":
        for table in TABLES:
            try:
                database.get_tables().index(table._meta.table_name)
            except ValueError:
                return False
    else:
        pass
    return True


if __name__ == "__main__":
    drop_tables()
    create_tables()



from pathlib import Path
import os
from typing import Optional, Union

from utils.style import color_green
from utils.ignore import IgnoreMatcher

def update_file(path: str, new_content: Union[str, bytes], project=None):
    """
    Update file with the new content.

    :param path: Full path to the file
    :param new_content: New content to write to the file
    :param project: Optional; a Project object related to the file update. Default is None.

    Any intermediate directories will be created if they don't exist.
    If file is text, it will be written using UTF-8 encoding.
    """
    # TODO: we should know where project root is and ensure no
    # files are written outside of it.
    os.makedirs(os.path.dirname(path), exist_ok=True)

    if isinstance(new_content, str):
        file_mode = "w"
        encoding = "utf-8"
    else:
        file_mode = "wb"
        encoding = None

    with open(path, file_mode, encoding=encoding) as file:
        file.write(new_content)
        if project is not None:  # project can be None only in tests
            if not project.skip_steps:
                print({"path": path, "line": None}, type='openFile')
            if not project.check_ipc():
                print(color_green(f"Updated file {path}"))


def get_file_contents(
    path: str, project_root_path: str
) -> dict[str, Union[str, bytes]]:
    """
    Get file content and metadata.

    :param path: Full path to the file
    :param project_root_path: Full path to the project root directory
    :return: Object with the following keys:
        - name: File name
        - path: Relative path to the file
        - content: File content (str or bytes)
        - full_path: Full path to the file

    If file is text, it will be read using UTF-8 encoding and `content`
    will be a Python string. If that fails, it will be treated as a
    binary file and `content` will be a Python bytes object.
    """
    # Normalize the path to avoid issues with different path separators
    full_path = os.path.normpath(path)

    try:
        # Assume it's a text file using UTF-8 encoding
        with open(full_path, "r", encoding="utf-8") as file:
            file_content = file.read()
    except UnicodeDecodeError:
        # If that fails, we'll treat it as a binary file
        with open(full_path, "rb") as file:
            file_content = file.read()
    except NotADirectoryError:
        raise ValueError(f"Path is not a directory: {path}")
    except FileNotFoundError:
        raise ValueError(f"File not found: {full_path}")
    except Exception as e:
        raise ValueError(f"Exception in get_file_contents: {e}")

    file_name = os.path.basename(path)
    relative_path = str(Path(path).parent.relative_to(project_root_path))

    if relative_path == '.':
        relative_path = ''

    return {
        "name": file_name,
        "path": relative_path,
        "content": file_content,
        "full_path": full_path,
    }


def get_directory_contents(
    directory: str,
    ignore: Optional[list[str]] = None,
) -> list[dict[str, Union[str, bytes]]]:
    """
    Get the content of all files in the given directory.

    :param directory: Full path to the directory to search
    :param ignore: List of files or folders to ignore (optional)
    :return: List of file objects as returned by `get_file_contents`

    See `get_file_contents()` for the details on the output structure
    and how files are read.
    """
    return_array = []

    matcher = IgnoreMatcher(ignore, root_path=directory)

    # TODO: Convert to use pathlib.Path.walk()
    for dpath, dirs, files in os.walk(directory):
        # In-place update of dirs so that os.walk() doesn't traverse them
        dirs[:] = [
            d for d in dirs
            if not matcher.ignore(os.path.join(dpath, d))
        ]

        for file in files:
            full_path = os.path.join(dpath, file)
            if matcher.ignore(full_path):
                continue

            return_array.append(get_file_contents(full_path, directory))

    return return_array


def clear_directory(directory: str, ignore: Optional[list[str]] = None):
    """
    Delete all files and directories (except ignored ones) in the given directory.

    :param dir_path: Full path to the directory to clear
    :param ignore: List of files or folders to ignore (optional)
    """
    matcher = IgnoreMatcher(ignore, root_path=directory)

    # TODO: Convert to use pathlib.Path.walk()
    for dpath, dirs, files in os.walk(directory, topdown=True):
        # In-place update of dirs so that os.walk() doesn't traverse them
        dirs[:] = [
            d for d in dirs
            if not matcher.ignore(os.path.join(dpath, d))
        ]

        for file in files:
            full_path = os.path.join(dpath, file)
            if matcher.ignore(full_path):
                continue

            os.remove(full_path)

        # Delete empty subdirectories not in ignore list
        for d in dirs:
            subdir_path = os.path.join(dpath, d)
            if not os.listdir(subdir_path):
                os.rmdir(subdir_path)



import json
import os
import yaml
from datetime import datetime
from dotenv import load_dotenv

load_dotenv()

USE_GPTPILOT_FOLDER = os.getenv('USE_GPTPILOT_FOLDER') == 'true'


# TODO: Parse files from the `.gpt-pilot` directory to resume a project - `user_stories` may have changed - include checksums for sections which may need to be reprocessed.
# TODO: Save a summary at the end of each task/sprint.
class DotGptPilot:
    """
    Manages the `.gpt-pilot` directory.
    """
    def __init__(self, log_chat_completions: bool = True):
        if not USE_GPTPILOT_FOLDER:
            return
        self.log_chat_completions = log_chat_completions
        self.dot_gpt_pilot_path = self.with_root_path('~', create=False)
        self.chat_log_path = self.chat_log_folder(None)

    def with_root_path(self, root_path: str, create=True):
        if not USE_GPTPILOT_FOLDER:
            return
        dot_gpt_pilot_path = os.path.expanduser(os.path.join(root_path, '.gpt-pilot'))
        self.dot_gpt_pilot_path = dot_gpt_pilot_path

        # Create the `.gpt-pilot` directory if required.
        if create and self.log_chat_completions:  # (... or ...):
            self.chat_log_folder(None)

        return dot_gpt_pilot_path

    def chat_log_folder(self, task):
        if not USE_GPTPILOT_FOLDER:
            return
        chat_log_path = os.path.join(self.dot_gpt_pilot_path, 'chat_log')
        if task is not None:
            chat_log_path = os.path.join(chat_log_path, 'task_' + str(task))

        os.makedirs(chat_log_path, exist_ok=True)
        self.chat_log_path = chat_log_path
        return chat_log_path

    def log_chat_completion(self, endpoint: str, model: str, req_type: str, messages: list[dict], response: str):
        if not USE_GPTPILOT_FOLDER:
            return
        if self.log_chat_completions:
            time = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
            with open(os.path.join(self.chat_log_path, f'{time}-{req_type}.yaml'), 'w', encoding="utf-8") as file:
                data = {
                    'endpoint': endpoint,
                    'model': model,
                    'messages': messages,
                    'response': response,
                }

                yaml.safe_dump(data, file, width=120, indent=2, default_flow_style=False, sort_keys=False)

    def log_chat_completion_json(self, endpoint: str, model: str, req_type: str, functions: dict, json_response: str):
        if not USE_GPTPILOT_FOLDER:
            return
        if self.log_chat_completions:
            time = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')

            with open(os.path.join(self.chat_log_path, f'{time}-{req_type}.json'), 'w', encoding="utf-8") as file:
                data = {
                    'endpoint': endpoint,
                    'model': model,
                    'functions': functions,
                    'response': json.loads(json_response),
                }

                json.dump(data, file, indent=2)

    def write_project(self, project):
        if not USE_GPTPILOT_FOLDER:
            return
        data = {
            'name': project.args['name'],
            'description': project.project_description,
            'user_stories': project.user_stories,
            'architecture': project.architecture,
            'system_dependencies': project.system_dependencies,
            'development_plan': project.development_plan,
        }

        with open(os.path.join(self.dot_gpt_pilot_path, 'project.yaml'), 'w') as file:
            yaml.safe_dump(data, file, width=120, indent=2, default_flow_style=False, sort_keys=False)



    # init CLI
    # 1. show the type of the app that needs to be created
    # 1.c ask user to press enter if it's ok, or to add the type of the app they want
        # if it's not ok, check if the wanted app CAN be created
            # if it can, print confirmation message and continue
            # if it can't, print error message and exit
    # 2. ask user for the main definition of the app
    # start the processing queue


# 2. show the user flow of the app
# 2.c ask user to press enter if it's ok, or to add the user flow they want
    # ask for input until they just press enter
    # recompute the user flow and ask again
# 3. show the COMPONENTS of the app
    # 3.1 frontend
    # 3.2 backend
    # 3.3 database
    # 3.4 config
    # 3.x ask user to press enter if it's ok, or to add the components they want
        # ask for input until they just press enter
        # recompute the components and ask again
# 4. break down the FILES that need to be created to support each of the components
    # ask user to press enter if it's ok, or to add the files they want
        # ask for input until they just press enter
        # recompute the files and ask again
# 5. loop through components (IMPORTANT!!!)
    # 5.1 loop through use cases
        # 5.1.1 for each case in each component, break down the files, functions and dependencies that need to be created
            # each function will have a description
            # in each loop, we will send all the previous files and functions so that LLM can change them if needed
# 6. break down the tests that need to be created
    # in the prompt, send all the files and functions
    # start from the high level tests and go down to the unit tests
    # 6.1 ask user to press enter if it's ok, or to add the tests they want
        # ask for input until they just press enter
        # recompute the tests and ask again
# 7. write the tests
# 8. write the files for each test
# 9. run each created test once the code is written
    # start from low level tests and do the high level tests last
    # track which test is related to which code
    # GPT should first say which functions will it use for a test and then we check if any of those functions is already written and if so, we send it to LLM to change it
    # track code coverage and increase to get to 100%
    # if the code requires something from config, ask the user to add it
    # if the code requires
    # when files overlap, ask LLM to combine them
# 10. try debugging 5 times
    # 10.1 if it doesn't work, ask the user to debug (!!!IMPORTANT!!!)
        # show them the explanations
        # ask for input if they want to enter something and retry 5 debugging attempts
# 11. create build/run script
# 12. RUN THE APP


# 4. show the components of the app setup
    # a. installation process
    # b. configuration process
    # c. running process
    # d. building process
    # e. testing process


# comments
# 1. Možemo koristiti dodatni model koji će izvlačiti iz GPT responsea što treba pokrenuti, što treba updateati, koji komentar složiti, etc. - da ne trebamo i to učiti originalni model in context



import platform
import questionary
import re
import sys
from database.database import save_user_input, get_saved_user_input
from utils.style import color_yellow_bold, style_config


def remove_ansi_codes(s: str) -> str:
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', s)


def styled_select(*args, **kwargs):
    kwargs["style"] = style_config.get_style()
    # TODO add saving and loading of user input
    return questionary.select(*args, **kwargs).unsafe_ask()  # .ask() is included here


def styled_text(project, question, ignore_user_input_count=False, style=None, hint=None):
    if not ignore_user_input_count:
        project.user_inputs_count += 1
        user_input = get_saved_user_input(project, question)
        if user_input is not None and user_input.user_input is not None and project.skip_steps:
            # if we do, use it
            project.checkpoints['last_user_input'] = user_input
            print(color_yellow_bold(f'Restoring user input id {user_input.id}: '), end='')
            print(color_yellow_bold(f'{user_input.user_input}'))
            return user_input.user_input

    if project.check_ipc():
        response = print(question, type='user_input_request')
        print(response)
    else:
        used_style = style if style is not None else style_config.get_style()
        question = remove_ansi_codes(question)  # Colorama and questionary are not compatible and styling doesn't work
        flush_input()
        response = questionary.text(question, style=used_style).unsafe_ask()  # .ask() is included here

    if not ignore_user_input_count:
        save_user_input(project, question, response, hint)

    print('\n\n', end='')
    return response


def get_user_feedback():
    return questionary.text('How did GPT Pilot do? Were you able to create any app that works? '
                            'Please write any feedback you have or just press ENTER to exit: ',
                            style=style_config.get_style()).unsafe_ask()


def ask_user_to_store_init_prompt():
    return questionary.text('We would appreciate if you let us store your initial app prompt. '
                            'If you are OK with that, please just press ENTER',
                            style=style_config.get_style()).unsafe_ask()


def flush_input():
    """Flush the input buffer, discarding all that's in the buffer."""
    try:
        if platform.system() == 'Windows':
            import msvcrt
            while msvcrt.kbhit():
                msvcrt.getch()
        else:
            import termios
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)
    except (ImportError, OSError):
        pass



import os.path
import re
from typing import Optional

from helpers.AgentConvo import AgentConvo
from helpers.Agent import Agent
from helpers.files import get_file_contents
from const.function_calls import GET_FILE_TO_MODIFY

from utils.exit import trace_code_event

class CodeMonkey(Agent):
    save_dev_steps = True

    def __init__(self, project, developer):
        super().__init__('code_monkey', project)
        self.developer = developer

    def get_original_file(self, code_changes_description, step, files):
        """
        Get the original file content and name.

        :param code_changes_description: description of the code changes
        :param step: information about the step being implemented
        :param files: list of files to send to the LLM
        :return: tuple of (file_name, file_content)
        """
        # If we're called as a result of debugging, we don't have the name/path of the file
        # to modify so we need to figure that out first.
        if 'path' not in step or 'name' not in step:
            file_to_change = self.identify_file_to_change(code_changes_description, files)
            step['path'] = os.path.dirname(file_to_change)
            step['name'] = os.path.basename(file_to_change)

        rel_path, abs_path = self.project.get_full_file_path(step['path'], step['name'])

        for f in files:
            # Take into account that step path might start with "/"
            if (f['path'] == step['path'] or (os.path.sep + f['path'] == step['path'])) and f['name'] == step['name'] and f['content']:
                file_content = f['content']
                break
        else:
            # If we didn't have the match (because of incorrect or double use of path separators or similar), fallback to directly loading the file
            try:
                file_content = get_file_contents(abs_path, self.project.root_path)['content']
                if isinstance(file_content, bytes):
                    # We should never want to change a binary file, but if we do end up here, let's not crash
                    file_content = "... <binary file, content omitted> ..."
            except ValueError:
                # File doesn't exist, we probably need to create a new one
                file_content = ""

        file_name = os.path.join(rel_path, step['name'])
        return file_name, file_content

    def implement_code_changes(
        self,
        convo: Optional[AgentConvo],
        code_changes_description: str,
        step: dict[str, str],
    ) -> AgentConvo:
        """
        Implement code changes described in `code_changes_description`.

        :param convo: AgentConvo instance (optional)
        :param task_description: description of the task
        :param code_changes_description: description of the code changes
        :param step: information about the step being implemented
        :param step_index: index of the step to implement
        """
        standalone = False
        if not convo:
            standalone = True
            convo = AgentConvo(self)

        files = self.project.get_all_coded_files()
        file_name, file_content = self.get_original_file(code_changes_description, step, files)
        content = file_content

        if file_content:
            # If we have the old file, try to replace individual code blocks
            replace_complete_file, content = self.replace_code_blocks(
                step,
                convo,
                standalone,
                code_changes_description,
                file_content,
                file_name,
                files,
            )
        else:
            # We're creating a new file, so we need to get full output
            replace_complete_file = True

        # If this is a new file or replacing individual code blocks failed,
        # replace the complete file.
        if replace_complete_file:
            content = self.replace_complete_file(
                convo,
                standalone,
                code_changes_description,
                file_content,
                file_name, files
            )

        if content and content != file_content:
            self.project.save_file({
                'path': step['path'],
                'name': step['name'],
                'content': content,
            })

        return convo

    def replace_code_blocks(
        self,
        step: dict[str, str],
        convo: AgentConvo,
        standalone: bool,
        code_changes_description: str,
        file_content: str,
        file_name: str,
        files: list[dict]
    ):
        llm_response = convo.send_message('development/implement_changes.prompt', {
            "full_output": False,
            "standalone": standalone,
            "code_changes_description": code_changes_description,
            "file_content": file_content,
            "file_name": file_name,
            "files": files,
        })

        replace_complete_file = False
        exchanged_messages = 2
        content = file_content

        # Allow for up to 2 retries
        while exchanged_messages < 7:
            if re.findall('(old|existing).+code', llm_response, re.IGNORECASE):
                trace_code_event("codemonkey-file-update-error", {
                    "error": "old-code-comment",
                    "llm_response": llm_response,
                })
                llm_response = convo.send_message('utils/llm_response_error.prompt', {
                    "error": (
                        "You must not omit any code from NEW_CODE. "
                        "Please don't use coments like `// .. existing code goes here`."
                    )
                })
                exchanged_messages += 2
                continue

            # Split the response into pairs of old and new code blocks
            block_pairs = self.get_code_blocks(llm_response)

            if len(block_pairs) == 0:
                if "```" in llm_response:
                    # We know some code blocks were outputted but we couldn't find them
                    print("Unable to parse code blocks from LLM response, asking to retry")
                    trace_code_event("codemonkey-file-update-error", {
                        "error": "error-parsing-blocks",
                        "llm_response": llm_response,
                    })

                    # If updating is more complicated than just replacing the complete file, don't bother.
                    if len(llm_response) > len(file_content):
                        replace_complete_file = True
                        break

                    llm_response = convo.send_message('utils/llm_response_error.prompt', {
                        "error": "I can't find CURRENT_CODE and NEW_CODE blocks in your response, please try again."
                    })
                    exchanged_messages += 2
                    continue
                else:
                    print(f"No changes required for {step['name']}")
                    break

            # Replace old code blocks with new code blocks
            errors = []
            for i, (old_code, new_code) in enumerate(block_pairs):
                try:
                    old_code, new_code = self.dedent(old_code, new_code)
                    content = self.replace(content, old_code, new_code)
                except ValueError as err:
                    errors.append((i + 1, str(err)))

            if not errors:
                break

            trace_code_event("codemonkey-file-update-error", {
                "error": "replace-errors",
                "llm_response": llm_response,
                "details": errors,
            })
            print(f"{len(errors)} error(s) while trying to update file, asking LLM to retry")

            if len(llm_response) > len(file_content):
                # If updating is more complicated than just replacing the complete file, don't bother.
                replace_complete_file = True
                break

            # Otherwise, identify the problem block(s) and ask the LLM to retry
            if content != file_content:
                error_text = (
                    "Some changes were applied, but these failed:\n" +
                    "\n".join(f"Error in change {i}:\n{err}" for i, err in errors) +
                    "\nPlease fix the errors and try again (only output the blocks that failed to update, not all of them)."
                )
            else:
                error_text = "\n".join(f"Error in change {i}:\n{err}" for i, err in errors)

            llm_response = convo.send_message('utils/llm_response_error.prompt', {
                "error": error_text,
            })
            exchanged_messages += 2
        else:
            # We failed after a few retries, so let's just replace the complete file
            print("Unable to modify file, asking LLM to output the complete new file")
            replace_complete_file = True

        if replace_complete_file:
            trace_code_event("codemonkey-file-update-error", {
                "error": "fallback-complete-replace",
                "llm_response": llm_response,
            })

        convo.remove_last_x_messages(exchanged_messages)
        return replace_complete_file, content

    def replace_complete_file(
        self,
        convo: AgentConvo,
        standalone: bool,
        code_changes_description: str,
        file_content: str,
        file_name: str,
        files: list[dict]
    ) -> str:
        """
        As a fallback, replace the complete file content.

        This should only be used if we've failed to replace individual code blocks.

        :param convo: AgentConvo instance
        :param standalone: True if this is a standalone conversation
        :param code_changes_description: description of the code changes
        :param file_content: content of the file being updated
        :param file_name: name of the file being updated
        :param files: list of files to send to the LLM
        :return: updated file content

        Note: if even this fails for any reason, the original content is returned instead.
        """
        llm_response = convo.send_message('development/implement_changes.prompt', {
            "full_output": True,
            "standalone": standalone,
            "code_changes_description": code_changes_description,
            "file_content": file_content,
            "file_name": file_name,
            "files": files,
        })
        start_pattern = re.compile(r"^\s*```([a-z0-9]+)?\n")
        end_pattern = re.compile(r"\n```\s*$")
        llm_response = start_pattern.sub("", llm_response)
        llm_response = end_pattern.sub("", llm_response)
        return llm_response


    def identify_file_to_change(self, code_changes_description: str, files: list[dict]) -> str:
        """
        Identify file to change based on the code changes description

        :param code_changes_description: description of the code changes
        :param files: list of files to send to the LLM
        :return: file to change
        """
        convo = AgentConvo(self)
        llm_response = convo.send_message('development/identify_files_to_change.prompt', {
            "code_changes_description": code_changes_description,
            "files": files,
        }, GET_FILE_TO_MODIFY)
        return llm_response["file"]

    @staticmethod
    def get_code_blocks(llm_response: str) -> list[tuple[str, str]]:
        """
        Split the response into code block(s).

        Ignores any content outside of code blocks.

        :param llm_response: response from the LLM
        :return: list of pairs of current and new blocks
        """
        pattern = re.compile(
            r"CURRENT_CODE:\n```([a-z0-9]+)?\n(.*?)\n```\nNEW_CODE:\n```([a-z0-9]+)?\n(.*?)\n```\nEND\s*",
            re.DOTALL
        )
        pairs = []
        for block in pattern.findall(llm_response):
            pairs.append((block[1], block[3]))
        return pairs

    @staticmethod
    def dedent(old_code: str, new_code: str) -> tuple[str, str]:
        """
        Remove common indentation from `old_code` and `new_code`.

        This is useful because the LLM will sometimes indent the code blocks MORE
        than in the original file, leading to no matches. Since we have indent
        compensation, we can just remove any extra indent as long as we do it
        consistently for both old and new code block.

        :param old_code: old code block
        :param new_code: new code block
        :return: tuple of (old_code, new_code) with common indentation removed
        """
        old_lines = old_code.splitlines()
        new_lines = new_code.splitlines()
        indent = 0
        while all(ol.startswith(" ") for ol in old_lines) and all(ol.startswith(" ") for ol in new_lines):
            indent -= 1
            old_lines = [ol[1:] for ol in old_lines]
            new_lines = [nl[1:] for nl in new_lines]
        return "\n".join(old_lines), "\n".join(new_lines)

    @staticmethod
    def replace(haystack: str, needle: str, replacement: str) -> str:
        """
        Replace `needle` text in `haystack`, allowing that `needle` is not
        indented the same as the matching part of `haystack` and
        compensating for it.

        :param haystack: text to search in
        :param needle: text to search for
        :param replacement: text to replace `needle` with
        :return: `haystack` with `needle` replaced with `replacement`

        Example:
        >>> haystack = "def foo():\n    pass"
        >>> needle = "pass"
        >>> replacement = "return 42"
        >>> replace(haystack, needle, replacement)
        "def foo():\n    return 42"

        If `needle` is not found in `haystack` even with indent compensation,
        or if it's found multiple times, raise a ValueError.
        """

        def indent_text(text: str, indent: int) -> str:
            return "\n".join((" " * indent + line) for line in text.splitlines())

        def indent_sensitive_match(haystack: str, needle: str) -> int:
            """
            Check if 'needle' is in 'haystack' but compare full lines.
            """
            # This is required so we don't match text "foo" (no indentation) with line "  foo"
            # (2 spaces indentation). We want exact matches so we know exact indentation needed.
            haystack_with_line_start_stop_markers = "\n".join(f"\x00{line}\x00" for line in haystack.splitlines())
            needle_with_line_start_stop_markers = "\n".join(f"\x00{line}\x00" for line in needle.splitlines())
            return haystack_with_line_start_stop_markers.count(needle_with_line_start_stop_markers)

        # Try from the largest indents to the smallest so that we know the correct indentation of
        # single-line old blocks that would otherwise match with 0 indent as well. If these single-line
        # old blocks were then replaced with multi-line blocks and indentation wasn't not correctly re-applied,
        # the new multiline block would only have the first line correctly indented. We want to avoid that.
        matching_old_blocks = []

        for indent in range(128, -1, -1):
            text = indent_text(needle, indent)
            if text not in haystack:
                # If there are empty lines in the old code, `indent_text` will indent them as well. The original
                # file might not have them indented as they're empty, so it is useful to try without indenting
                # those empty lines.
                text = "\n".join(
                    (line if line.strip() else "")
                    for line
                    in text.splitlines()
                )
            n_matches = indent_sensitive_match(haystack, text)
            for i in range(n_matches):
                matching_old_blocks.append((indent, text))

        if len(matching_old_blocks) == 0:
            raise ValueError(
                f"Old code block not found in the original file:\n```\n{needle}\n```\n"
                "Old block *MUST* contain the exact same text (including indentation, empty lines, etc.) as the original file "
                "in order to match."
            )

        if len(matching_old_blocks) > 1:
            raise ValueError(
                f"Old code block found more than once ({len(matching_old_blocks)} matches) in the original file:\n```\n{needle}\n```\n\n"
                "Please provide larger blocks (more context) to uniquely identify the code that needs to be changed."
            )

        indent, text = matching_old_blocks[0]
        indented_replacement = indent_text(replacement, indent)
        return haystack.replace(text, indented_replacement)



import os

DATABASE_TYPE = os.getenv("DATABASE_TYPE", "sqlite")
DB_NAME = os.getenv("DB_NAME")
DB_HOST = os.getenv("DB_HOST")
DB_PORT = os.getenv("DB_PORT")
DB_USER = os.getenv("DB_USER")
DB_PASSWORD = os.getenv("DB_PASSWORD")



import json
from uuid import uuid4

from utils.telemetry import telemetry
from utils.exit import trace_code_event
from const.telemetry import LOOP_THRESHOLD


class Task:
    """
    Task data structure to store information about the current task. The task data structure is sent to telemetry.
    Currently used to trace big loops in the code.

    >>> from utils.task import Task

    To set up a new task:

    >>> task = Task()

    To set a value:

    >>> task.set('task_description', 'test')

    To increment a value:

    >>> task.inc('steps')

    To start a new task:

    >>> task.start_new_task('test', 1)

    When debugging recursion happens inside a task (see pilot/helpers/Debugger.py) we add a debugging task to the
    task data structure. To add a debugging task:

    >>> task.add_debugging_task(1, {'command': 'test'}, 'This is not working', 'Command is not working')

    To clear the task:

    >>> task.clear()

    To send the task:

    >>> task.send()

    Note: the task will be sent automatically if the number of steps exceeds the threshold
    """

    def __init__(self):
        self.initial_data = {
            'task_description': '',
            'task_number': 0,
            'steps': 0,
            'debugging': [],
        }
        self.data = self.initial_data.copy()
        self.ping_extension = True

    def set(self, key: str, value: any):
        """
        Set a value in the task data

        :param key: key to set
        :param value: value to set
        """
        self.data[key] = value

    def inc(self, key: str, value: int = 1):
        """
        Increment a value in the task data

        :param key: key to increment
        :param value: value to increment by
        """
        self.data[key] += value
        if key == 'steps' and self.data[key] == LOOP_THRESHOLD + 1:
            self.send()

    def start_new_task(self, task_description: str, i: int):
        """
        Start a new task

        :param task_description: description of the task
        :param i: task number
        """
        self.send(name='loop-end')
        self.clear()
        self.set('task_description', task_description)
        self.set('task_number', i)
        self.set('loopId', f"{uuid4()}")

    def add_debugging_task(self, recursion_layer: int = None,
                           command: dict = None,
                           user_input: str = None,
                           issue_description: str = None):
        """
        Add a debugging task to the task data structure

        :param recursion_layer: recursion layer
        :param command: command to debug
        :param user_input: user input
        :param issue_description: description of the issue
        """
        self.data['debugging'].append({
            'recursion_layer': recursion_layer,
            'command': command,
            'user_inputs': [user_input] if user_input is not None else [],
            'issue_description': issue_description,
        })

    def add_user_input_to_debugging_task(self, user_input: str):
        """
        Add user input to the last debugging task

        :param user_input: user input
        """
        if self.data.get('debugging') and len(self.data['debugging']) > 0:
            self.data['debugging'][-1]['user_inputs'].append(user_input)

    def clear(self):
        """
        Clear all the task data
        """
        self.data = self.initial_data.copy()

    def send(self, name: str = 'loop-start'):
        """
        Send the task data to telemetry

        :param name: name of the event
        """
        if self.data['steps'] > LOOP_THRESHOLD:
            full_data = telemetry.data.copy()
            full_data['task_with_loop'] = self.data.copy()
            trace_code_event(name=name, data=full_data)
            if self.ping_extension:
                print(json.dumps({
                    'pathId': telemetry.telemetry_id,
                    'data': full_data,
                }), type='loopTrigger')
                # TODO: see if we want to ping the extension multiple times
                self.ping_extension = False

    def exit(self):
        """
        Send the task data to telemetry and exit the process
        """
        self.send(name='loop-end')



MESSAGE_TYPE = {
    'verbose': 'verbose',
    'stream': 'stream',
    'user_input_request': 'user_input_request',   # Displayed above the
    'hint': 'hint',                        # Hint text, eg "Do you want to add anything else? If not, just press ENTER."
    'info': 'info',                        # JSON data can be sent to progress `progress_stage`
    'local': 'local',
    'run_command': 'run_command',          # Command to run server needed for extension only
    'project_folder_name': 'project_folder_name',  # Project folder name for extension only
    'button': 'button',                    # Button text for extension only
    'buttons-only': 'buttons-only',        # Button text for extension only but input field is disabled
    'exit': 'exit',                        # Exit message to let extension know we are done
    'ipc': 'ipc',                          # Regular print message that is for extension only
    'openFile': 'openFile',                # Open a file in extension
    'loadingFinished': 'loadingFinished',  # Marks end of loading project
    'loopTrigger': 'loopTrigger',          # Trigger loop feedback popup in extension
}

LOCAL_IGNORE_MESSAGE_TYPES = [
    'info',
    'project_folder_name',
    'button',
    'buttons-only',
    'exit',
    'ipc',
    'openFile',
    'loadingFinished',
    'loopTrigger',
]


from peewee import ForeignKeyField, CharField, BooleanField, DateTimeField
from database.config import DATABASE_TYPE
from database.models.components.base_models import BaseModel
from database.models.app import App
from database.models.development_steps import DevelopmentSteps
from database.models.components.sqlite_middlewares import JSONField
from playhouse.postgres_ext import BinaryJSONField


class Feature(BaseModel):
    app = ForeignKeyField(App, backref='feature', on_delete='CASCADE')
    summary = CharField()

    if DATABASE_TYPE == 'postgres':
        messages = BinaryJSONField(null=True)
    else:
        messages = JSONField(null=True)

    previous_step = ForeignKeyField(DevelopmentSteps, column_name='previous_step')
    completed = BooleanField(default=False)
    completed_at = DateTimeField(null=True)



from peewee import AutoField, CharField, ForeignKeyField

from database.models.components.base_models import BaseModel
from database.models.app import App
from database.models.user import User


class UserApps(BaseModel):
    id = AutoField()
    app = ForeignKeyField(App, on_delete='CASCADE')
    user = ForeignKeyField(User, on_delete='CASCADE')
    workspace = CharField(null=True)

    class Meta:
        table_name = 'user_apps'
        indexes = (
            (('app', 'user'), True),
        )



from setuptools import setup, find_packages

VERSION = "0.0.10"

requirements = open("requirements.txt").readlines()

if __name__ == "__main__":
    setup(
        name='gpt-pilot',
        version=VERSION,
        packages=find_packages(),
        url='https://www.pythagora.ai',
        license='MIT',
        author_email='info@pythagora.ai',
        description='GPT Pilot - an AI developer that works with you to build complex projects',
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
        ],
        install_requires=requirements,
        python_requires='>=3.9',
    )



import unittest
from pilot.utils.style import style_config, Theme, ColorName, get_color_function


class TestColorStyle(unittest.TestCase):
    def test_initialization(self):
        print("\n[INFO] Testing Theme Initialization...")
        style_config.set_theme(Theme.DARK)
        print(f"[INFO] Set theme to: {Theme.DARK}, Current theme: {style_config.theme}")
        self.assertEqual(style_config.theme, Theme.DARK)

        style_config.set_theme(Theme.LIGHT)
        print(f"[INFO] Set theme to: {Theme.LIGHT}, Current theme: {style_config.theme}")
        self.assertEqual(style_config.theme, Theme.LIGHT)

    def test_color_function(self):
        dark_color_codes = {
            ColorName.RED: "\x1b[31m",
            ColorName.GREEN: "\x1b[32m",
            # ... other colors
        }
        light_color_codes = {
            ColorName.RED: "\x1b[91m",
            ColorName.GREEN: "\x1b[92m",
            # ... other colors
        }
        reset = "\x1b[0m"

        # Test DARK theme
        print("\n[INFO] Testing DARK Theme Colors...")
        style_config.set_theme(Theme.DARK)
        for color_name, code in dark_color_codes.items():
            with self.subTest(color=color_name):
                color_func = get_color_function(color_name, bold=False)
                print(f"[INFO] Testing color: {color_name}, Expect: {code}Test, Got: {color_func('Test')}")
                self.assertEqual(color_func("Test"), f"{code}Test{reset}")

                color_func = get_color_function(color_name, bold=True)
                print(
                    f"[INFO] Testing color (bold): {color_name}, Expect: {code}\x1b[1mTest, Got: {color_func('Test')}")
                self.assertEqual(color_func("Test"), f"{code}\x1b[1mTest{reset}")

        # Test LIGHT theme
        print("\n[INFO] Testing LIGHT Theme Colors...")
        style_config.set_theme(Theme.LIGHT)
        for color_name, code in light_color_codes.items():
            with self.subTest(color=color_name):
                color_func = get_color_function(color_name, bold=False)
                print(f"[INFO] Testing color: {color_name}, Expect: {code}Test, Got: {color_func('Test')}")
                self.assertEqual(color_func("Test"), f"{code}Test{reset}")

                color_func = get_color_function(color_name, bold=True)
                print(
                    f"[INFO] Testing color (bold): {color_name}, Expect: {code}\x1b[1mTest, Got: {color_func('Test')}")
                self.assertEqual(color_func("Test"), f"{code}\x1b[1mTest{reset}")



import logging

from peewee import ForeignKeyField, BlobField

from database.models.components.base_models import BaseModel
from database.models.development_steps import DevelopmentSteps
from database.models.app import App
from database.models.files import File

log = logging.getLogger(__name__)


class SmartBlobField(BlobField):
    """
    A binary blob field that can also accept/return utf-8 strings.

    This is a temporary workaround for the fact that we're passing either binary
    or string contents to the database. Once this is cleaned up, we should only
    accept binary content and explcitily convert from/to strings as needed.
    """

    def db_value(self, value):
        if isinstance(value, str):
            log.warning("FileSnapshot content is a string, expected bytes, working around it.")
            value = value.encode("utf-8")
        return super().db_value(value)

    def python_value(self, value):
        val = bytes(super().python_value(value))
        try:
            return val.decode("utf-8")
        except UnicodeDecodeError:
            return val


class FileSnapshot(BaseModel):
    app = ForeignKeyField(App, on_delete='CASCADE')
    development_step = ForeignKeyField(DevelopmentSteps, backref='files', on_delete='CASCADE')
    file = ForeignKeyField(File, on_delete='CASCADE', null=True)
    content = SmartBlobField()

    class Meta:
        table_name = 'file_snapshot'
        indexes = (
            (('development_step', 'file'), True),
        )



import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest.mock import patch, call

import pytest

from pilot.helpers.files import get_file_contents, get_directory_contents, update_file


@patch("pilot.helpers.files.open")
@patch("pilot.helpers.files.os")
def test_update_file_creates_directories(mock_os, mock_open):
    """Test that update_file creates intermediate directories if they don't exist."""

    mock_os.path.dirname = os.path.dirname
    update_file("/path/to/file", "content")
    mock_os.makedirs.assert_called_once_with("/path/to", exist_ok=True)


@patch("pilot.helpers.files.open")
@patch("pilot.helpers.files.os")
def test_update_file_creates_text_file(mock_os, mock_open):
    """Test that update_file creates intermediate directories if they don't exist."""

    update_file("/path/to/file", "無為")
    mock_open.assert_called_once_with("/path/to/file", "w", encoding="utf-8")
    mock_open.return_value.__enter__.return_value.write.assert_called_once_with("無為")


@patch("pilot.helpers.files.open")
@patch("pilot.helpers.files.os")
def test_update_file_creates_binary_file(mock_os, mock_open):
    """Test that update_file creates intermediate directories if they don't exist."""

    update_file("/path/to/file", b"\x00\x00\x00")
    mock_open.assert_called_once_with("/path/to/file", "wb", encoding=None)
    mock_open.return_value.__enter__.return_value.write.assert_called_once_with(b"\x00\x00\x00")


@pytest.mark.parametrize(
    ("source", "expected_encoded"),
    [
        ("file.txt", b"file.txt"),
        ("foo.txt - 無為", b"foo.txt - \xe7\x84\xa1\xe7\x82\xba"),
        (b"\xff\xff\xff", b"\xff\xff\xff"),
    ],
)
def test_update_file_with_encoded_content(source, expected_encoded):
    # Can't use NamedTemporaryFile this as a context manager because Windows
    # doesn't allow O_TEMPORARY files (with delete=True) to be opened
    # twice, defeating the purpose.
    file = NamedTemporaryFile(delete=False)
    update_file(file.name, source)
    assert file.read() == expected_encoded

    file.close()
    os.remove(file.name)

@pytest.mark.parametrize(
    ("encoded", "expected"),
    [
        (b"file.txt", "file.txt"),
        (b"foo.txt - \xe7\x84\xa1\xe7\x82\xba", "foo.txt - 無為"),
        (b"\xff\xff\xff", b"\xff\xff\xff"),
    ],
)
def test_get_file_contents(encoded, expected):
    file = NamedTemporaryFile(delete=False)
    file.write(encoded)
    file.flush()

    file_path = Path(file.name)
    data = get_file_contents(file.name, file_path.anchor)
    assert data == {
        "content": expected,
        "name": file_path.name,
        "path": str(file_path.parent.relative_to(file_path.anchor)),
        "full_path": file.name,
    }
    file.close()
    os.remove(file.name)

@patch("pilot.helpers.files.open")
@patch("pilot.helpers.files.os")
def test_get_directory_contents_mocked(mock_os, mock_open):
    """
    Test that get_directory_contents traverses the directory tree,
    ignores specified ignore files/folders, and can handle both
    text and binary files.
    """

    # Normalize path to use correct path separators
    def np(path: str) -> str:
        return str(Path(path))

    mock_os.path.join = os.path.join
    mock_os.path.normpath = os.path.normpath
    mock_os.path.basename = os.path.basename

    mock_walk = mock_os.walk
    mock_walk.return_value = [
        (np("/fake/root"), ["foo", "to-ignore", "bar"], ["file.txt", "to-ignore.txt"]),
        (np("/fake/root/foo"), [], ["foo.txt"]),
        (np("/fake/root/bar"), [], ["bar.txt"]),
    ]
    mock_open.return_value.__enter__.return_value.read.side_effect = [
        "file.txt",
        "foo.txt - 無為",
        UnicodeDecodeError("utf-8", b"\xff\xff\xff", 0, 1, "invalid start byte"),
        b"\xff\xff\xff",
    ]

    data = get_directory_contents(np("/fake/root"), ["to-ignore", "to-ignore.txt"])
    assert data == [
        {
            "content": "file.txt",
            "full_path": np("/fake/root/file.txt"),
            "name": "file.txt",
            "path": "",
        },
        {
            "content": "foo.txt - 無為",
            "full_path": np("/fake/root/foo/foo.txt"),
            "name": "foo.txt",
            "path": "foo",
        },
        {
            "content": b"\xff\xff\xff",
            "full_path": np("/fake/root/bar/bar.txt"),
            "name": "bar.txt",
            "path": "bar",
        },
    ]
    mock_walk.assert_called_once_with(np("/fake/root"))


def test_get_directory_contents_live():
    files = get_directory_contents(
        os.path.dirname(os.path.dirname(__file__)), [".pytest_cache", "agents", "__init__.py"]
    )

    # Check this file was loaded as a text file
    this_file = [f for f in files if f["name"] == "test_files.py"][0]
    assert this_file["path"] == "helpers"
    assert this_file["full_path"] == __file__
    assert isinstance(this_file["content"], str)
    assert "test_get_directory_contents_live()" in this_file["content"]

    # Check that the binary file was ignored
    image_files = [
        f
        for f in files
        if f["path"] == "helpers" and f["name"] == "testlogo.png"
    ]
    assert image_files == []

    # Check that the ignore list works
    assert all(file["name"] != "__init__.py" for file in files)
    assert any(file["path"] == "database" for file in files)



from yaspin import yaspin
from yaspin.spinners import Spinners


def spinner_start(text="Processing..."):
    spinner = yaspin(Spinners.line, text=text)
    spinner.start()
    return spinner


def spinner_stop(spinner):
    if spinner is not None:
        spinner.stop()



from database.config import DATABASE_TYPE
from database.models.components.progress_step import ProgressStep
from database.models.components.sqlite_middlewares import JSONField
from playhouse.postgres_ext import BinaryJSONField


class UserStories(ProgressStep):
    if DATABASE_TYPE == 'postgres':
        user_stories = BinaryJSONField()
    else:
        user_stories = JSONField()  # Custom JSON field for SQLite
    class Meta:
        table_name = 'user_stories'



import os
from pathlib import Path
from database.database import save_user_app


def get_parent_folder(folder_name):
    current_path = Path(os.path.abspath(__file__))  # get the path of the current script

    while current_path.name != folder_name:  # while the current folder name is not 'folder_name'
        current_path = current_path.parent  # go up one level

    return current_path.parent


def setup_workspace(args) -> str:
    """
    Creates & returns the path to the project workspace.
    Also creates a 'tests' folder inside the workspace.
    :param args: may contain 'root' key
    """
    workspace = args.get('workspace')
    if workspace:
        project_path = workspace
    else:
        root = args.get('root') or get_parent_folder('pilot')
        name = args.get('name', 'default_project_name')
        project_path = create_directory(os.path.join(root, 'workspace'), name)

    create_directory(project_path, 'tests')
    try:
        save_user_app(args.get('user_id'), args.get('app_id'), project_path)
    except Exception as e:
        print(f'Error saving user app: {str(e)}')

    print(os.path.basename(project_path), type='project_folder_name')
    return project_path


def create_directory(parent_directory, new_directory):
    new_directory_path = os.path.join(parent_directory, new_directory)
    os.makedirs(new_directory_path, exist_ok=True)

    return new_directory_path


def count_lines_of_code(files):
    return sum(len(file['content'].splitlines()) for file in files)
